function [gm,Mm,lm,Plyap,cvx_status] = wcgLPVAnalysis(PICT,PSIICT,IQCMode)
%%WCGANALYSIS returns the worst-case gain using IQC-based stability analysis.
% function [gm,Mm,lm,Plyap,cvx_status] = wcgAnalysis(PICT,PSIICT,IQCMode)
% Input arguments:
% (1) PICT is the interconnection of the nominal plant and psi filters.
% (2) PSIICT is the interconnection of the PSI filters.
% (3) IQCMode defines the synthesis mode.
% Output arguments:
% (1) gm is the L2-L2 norm of lft(DELTA,G).
% (2) Mm is the IQC multiplier in the time domain (single matrix).
% (3) lm are the IQC multipliers in the time domain (multiple lambdas).
% (4) Plyap is the Lyapunov function matrix for the plant-filter ICT.
% (5) cvx_status is the status of the optimization.

%% Set generalized errors identifier string
if isfield(PICT.OutputGroup,'Measurement')
    GenErrorStr = 'Measurement';
elseif isfield(PICT.OutputGroup,'EstimationError')
    GenErrorStr = 'EstimationError';
else
    fprintf('\n - The output groups of the argument to wcgAnalysis are:\n');
    PICT.OutputGroup
    error('Unknown generalized error output name.');
end

%% Extract SS matrices
A = PICT.a;
B1 = PICT(:,'Uncertain').b;
B2 = PICT(:,'Disturbance').b;
C1 = PICT({'Uncertainv','Uncertainw'},:).c; % SingleM mode
C2 = PICT(GenErrorStr,:).c;
D11 = PICT({'Uncertainv','Uncertainw'},'Uncertain').d; % SingleM mode
D12 = PICT({'Uncertainv','Uncertainw'},'Disturbance').d; % SingleM mode
D21 = PICT(GenErrorStr,'Uncertain').d;
D22 = PICT(GenErrorStr,'Disturbance').d;

%% Extract matrix sizes
n = length(PICT.StateName);
nw = length(PICT.InputGroup.Uncertain);
m = length(PICT.InputGroup.Disturbance);
nv = length(PICT.OutputGroup.UncertainF1) - nw; % special treatment to restore nv
p = length(PICT.OutputGroup.(GenErrorStr));

% Error protection
if nv ~= nw
    warning('I/O sizes of DELTA do not match: nv != nw.');
end

% Number of IQC multipiers
NIdx = strfind(fieldnames(PICT.OutputGroup),'UncertainF');
N = sum([NIdx{:}]);

% Number of grid points in PSS object (currently only supports 1D grid)
ng = size(PICT.Parameter.rho.GridData,1); % number of grid points

%% Preparing C1s, D11s, and D12s matrices for MultipleLm mode
% Placeholders for cell arrays for C1s, D11s, D12s
C1s = cell(N,1);
D11s = cell(N,1);
D12s = cell(N,1);

% Assign values to C1s{k,1}, D11s{k,1}, D12s{k,1}
for k = 1:N % sweep over number of IQC multipliers
    C1s{k,1} = PICT(['UncertainF',num2str(k)],:).c; % MultipleLm mode
    D11s{k,1} = PICT(['UncertainF',num2str(k)],'Uncertain').d; % MultipleLm mode
    D12s{k,1} = PICT(['UncertainF',num2str(k)],'Disturbance').d; % MultipleLm mode
end

% Definition of generic M for MultipleLm mode
M = [eye(nv), zeros(nv,nw); zeros(nw,nv), -eye(nw)];

%% Extracting (1,1) block from PSIICT
PSI11 = ss(PSIICT('Uncertainv','v')); % extract the (1,1) block of PSIICT
[Apsi,Bpsi,Cpsi,Dpsi] = ssdata(PSI11); % extract ss data of PSI11
nstate_psi = size(Apsi,1); % number of states in PSI11
ninput_psi = size(Dpsi,2); % number of inputs in PSI11
noutput_psi = size(Dpsi,1); % number of outputs in PSI11

%% Analysis condition: gain minimization
ep = 1e-8; % small positive number

% Set Uflag = 1 to enforce U>=0 but not KYP_R>=0
% Set Uflag = 0 to enforce KYP_R>=0 but not U>=0
Uflag = 1;

cvx_begin sdp quiet
% Definition of variables in LMI
    variable Plyap(n,n) symmetric
    variable lm(N,1) nonnegative
    variable Pm(N*nv,N*nv) symmetric
    if N*nv == 1 % if Rm is a scalar
        variable Rm(N*nv,N*nv) % cvx can't define a scalar as skew_symmetric
    else % if Rm is a matrix
        variable Rm(N*nv,N*nv) skew_symmetric
    end
    variable gm
    if nstate_psi > 0 && ~Uflag
        variable Y(nstate_psi,nstate_psi) symmetric
    end
    
% Definition of terms in LMI
KYP_SYM = cell(ng,1);

for kk = 1:ng
    % Extract data at a grid point
    Ap = A.Data(:,:,kk);
    B1p = B1.Data(:,:,kk);
    B2p = B2.Data(:,:,kk);
    C1p = C1.Data(:,:,kk);
    C2p = C2.Data(:,:,kk);
    D11p = D11.Data(:,:,kk);
    D12p = D12.Data(:,:,kk);
    D21p = D21.Data(:,:,kk);
    D22p = D22.Data(:,:,kk);

    % T1
    T1 = ...
    [Ap'*Plyap + Plyap*Ap, Plyap*B1p, Plyap*B2p, C2p';...
     B1p'*Plyap, zeros(nw,nw), zeros(nw,m), D21p';...
     B2p'*Plyap, zeros(m,nw), -gm.*eye(m), D22p';...
     C2p, D21p, D22p, -gm.*eye(p)];
 
     % T2
     switch IQCMode
         case 'SingleM'
             Mm = [Pm,Rm;Rm',-Pm]; % as given in Scherer_08Automatica
             T2 = [([C1p, D11p, D12p]'*...
                 Mm*...
                 [C1p, D11p, D12p]),...
                 zeros(n+nw+m,p);...
                 zeros(p,n+nw+m), zeros(p,p)];
         case 'MultipleLm'
             T2 = 0; % placeholder for term
             for k = 1:N % sweep over number of IQC multipliers
                 T2 = T2 + ...
                     lm(k,1).*...
                     [([C1s{k,1}, D11s{k,1}, D12s{k,1}]'*...
                     M*...
                     [C1s{k,1}, D11s{k,1}, D12s{k,1}]),...
                     zeros(n+nw+m,p);...
                     zeros(p,n+nw+m), zeros(p,p)];
             end
         otherwise
             error(['Analysis mode IQCMode = ',IQCMode,' is invalid.']);
     end
    
    % Force symmetry in KYP LMI
    KYP_ORG = T1 + T2; % original KYP LMI
    KYP_SYM{kk,1} = 0.5*(KYP_ORG + KYP_ORG'); % symmetry forced
    sym_err = norm(cvx_basis(0.5*(KYP_ORG-KYP_SYM{kk,1})),'fro')...
        /norm(cvx_basis(KYP_ORG),'fro');
%     fprintf('\n - Forcing symmetry of KYP LMI. Relative error is %1.4e.\n',...
%         sym_err);
end

    % Enforce positive definiteness of psi'*Pm*psi on jw-axis.
    % See Scherer_08AUT, section 5, "Numerical example" for details
    if nstate_psi > 0 && ~Uflag
        KYP_psi = [Apsi'*Y + Y*Apsi,Y*Bpsi;Bpsi'*Y,zeros(ninput_psi)] + ...
            [Cpsi,Dpsi]'*Pm*[Cpsi,Dpsi];
    else
        KYP_psi = Pm;
    end

% Gain minimization subject to LMI constraints
    minimize gm
    subject to
%         Plyap >= ep.*eye(n);
    for kk = 1:ng
        KYP_SYM{kk,1} <= -ep.*eye(n+nw+m+p); % KYP LMI
    end
        gm >= ep; % gm > 0
        switch IQCMode
            case 'SingleM'
                if N*nv == 1
                    Rm == 0; % force Rm == 0 when Rm is a scalar
                end
                lm == zeros(N,1); % set lambdas to zero
            case 'MultipleLm'
                Pm == zeros(N*nv,N*nv); % set Pm to zero
                Rm == zeros(N*nv,N*nv); % set Rm to zero
                Mm = [Pm,Rm;Rm',-Pm]; % return Mm
            otherwise
                error(['Analysis mode IQCMode = ',IQCMode,' is invalid.']);
        end
        if nstate_psi > 0 && ~Uflag
            KYP_psi >= ep.*eye(nstate_psi+ninput_psi); % KYP LMI for PSI11
        else
            KYP_psi >= ep.*eye(noutput_psi); % KYP LMI for PSI11
        end        
cvx_end

%% Error protection
switch cvx_status
    case 'Infeasible'
        error(['CVX status is: ',cvx_status,'.']);
    case {'Inaccurate/Solved','Failed'}
        warning(['CVX status is: ',cvx_status,'.']);
end

end
